#include <linux/delay.h>
#include <linux/module.h>
#include <linux/input.h>
#include <linux/serio.h>
#include <linux/libps2.h>
#include <linux/i8042.h>
#include "psmouse.h"
#include "elantech.h"

#define elantech_debug(format, arg...)				\
	do {							\
		if (etd->debug)					\
			printk(KERN_DEBUG format, ##arg);	\
	} while (0)

static int EF_023_DEBUG;
static int Public_ETP_YMAX_V2;
static int Public_ETP_2FT_YMAX;

/*
 * Send a Synaptics style sliced query command
 */
static int synaptics_send_cmd(struct psmouse *psmouse, unsigned char c,
				unsigned char *param)
{
	if (psmouse_sliced_command(psmouse, c) ||
	    ps2_command(&psmouse->ps2dev, param, PSMOUSE_CMD_GETINFO)) {
		pr_err("elantech.c: synaptics_send_cmd query 0x%02x failed.\n", c);
		return -1;
	}

	return 0;
}

/*
 * A retrying version of ps2_command
 */
static int elantech_ps2_command(struct psmouse *psmouse,
				unsigned char *param, int command)
{
	struct ps2dev *ps2dev = &psmouse->ps2dev;
	struct elantech_data *etd = psmouse->private;
	int rc;
	int tries = ETP_PS2_COMMAND_TRIES;

	do {
		rc = ps2_command(ps2dev, param, command);
		if (rc == 0)
			break;
		tries--;
		elantech_debug("elantech.c: retrying ps2 command 0x%02x (%d).\n",
			command, tries);
		msleep(ETP_PS2_COMMAND_DELAY);
	} while (tries > 0);

	if (rc)
		pr_err("elantech.c: ps2 command 0x%02x failed.\n", command);

	return rc;
}

/*
 * Send an Elantech style special command to read a value from a register
 */
static int elantech_read_reg(struct psmouse *psmouse, unsigned char reg,
				unsigned char *val)
{
	struct elantech_data *etd = psmouse->private;
	unsigned char param[3];
	int rc = 0;

	if (reg < 0x10 || reg > 0x26)
		return -1;

	if (reg > 0x11 && reg < 0x20)
		return -1;

	switch (etd->hw_version) {
	case ETF020022:
	case ETF020600:
		if (psmouse_sliced_command(psmouse, ETP_REGISTER_READ) ||
		    psmouse_sliced_command(psmouse, reg) ||
		    ps2_command(&psmouse->ps2dev, param, PSMOUSE_CMD_GETINFO)) {
			rc = -1;
		}
		break;

	case ETF020030:
	case ETF0208:
	case ETF020B00:
	case ETF0402:
	case ETF0401:
	case ETF0403:
		if (elantech_ps2_command(psmouse,  NULL, ETP_PS2_CUSTOM_COMMAND) ||
		    elantech_ps2_command(psmouse,  NULL, ETP_REGISTER_READ) ||
		    elantech_ps2_command(psmouse,  NULL, ETP_PS2_CUSTOM_COMMAND) ||
		    elantech_ps2_command(psmouse,  NULL, reg) ||
		    elantech_ps2_command(psmouse, param, PSMOUSE_CMD_GETINFO)) {
			rc = -1;
		}
		break;
	}

	if (rc)
		pr_err("elantech.c: failed to read register 0x%02x.\n", reg);
	else
		*val = param[0];

	return rc;
}

/*
 * Send an Elantech style special command to write a register with a value
 */
static int elantech_write_reg(struct psmouse *psmouse, unsigned char reg,
				unsigned char val)
{
	struct elantech_data *etd = psmouse->private;
	int rc = 0;
	if(EF_023_DEBUG == 0) {
		if (reg < 0x10 || reg > 0x26)
			return -1;

		if (reg > 0x11 && reg < 0x20)
			return -1;
	}

	switch (etd->hw_version) {
	case ETF020022:
	case ETF020600:
		if (psmouse_sliced_command(psmouse, ETP_REGISTER_WRITE) ||
		    psmouse_sliced_command(psmouse, reg) ||
		    psmouse_sliced_command(psmouse, val) ||
		    ps2_command(&psmouse->ps2dev, NULL, PSMOUSE_CMD_SETSCALE11)) {
			rc = -1;
		}
		break;

	case ETF020030:
	case ETF0208:
	case ETF020B00:
	case ETF0402:
	case ETF0401:
	case ETF0403:
		if (elantech_ps2_command(psmouse, NULL, ETP_PS2_CUSTOM_COMMAND) ||
		    elantech_ps2_command(psmouse, NULL, ETP_REGISTER_WRITE) ||
		    elantech_ps2_command(psmouse, NULL, ETP_PS2_CUSTOM_COMMAND) ||
		    elantech_ps2_command(psmouse, NULL, reg) ||
		    elantech_ps2_command(psmouse, NULL, ETP_PS2_CUSTOM_COMMAND) ||
		    elantech_ps2_command(psmouse, NULL, val) ||
		    elantech_ps2_command(psmouse, NULL, PSMOUSE_CMD_SETSCALE11)) {
			rc = -1;
		}
		break;
	}

	if (rc)
		pr_err("elantech.c: failed to write register 0x%02x with value 0x%02x.\n",
			reg, val);

	return rc;
}


static int ETF0208_DEBUG_SOLUTION(int fingers,int button,int w)
{
	static int can_bug=0,debug_condition_1=0,count_report=0,debug_condition_3=0,debug_condition_5=0;
	static int solution=0;
	
	if(solution==3)
		return 0;

	if(fingers >=2 || (fingers == 1 && button > 0))
		can_bug=1;
			
		
	if(can_bug && button > 0)
		debug_condition_1=1;

	if(fingers <= 1 && can_bug && button==0)
		can_bug=0;
	
	if(debug_condition_1&& fingers ==0 && button > 0){
		count_report++;
		//printk(KERN_DEBUG "2 count_report=%d \n", count_report);
		if(count_report==4){
			debug_condition_1=0;
			EF_023_DEBUG=2;
			solution=2;
			return 1;
			
		}

	}
	else if(debug_condition_1  && fingers == 1 && button > 0){
		count_report=0;			
	}
	else if(debug_condition_1 &&  count_report > 0 &&  count_report < 4  && button == 0){
		EF_023_DEBUG=1;
		solution=1;
		debug_condition_1=0;
		count_report=0;
		return 1;
	}
	else if(fingers ==0 && button == 0){
		count_report=0;	
		debug_condition_1=0;
	}
	
	if(debug_condition_5==1 &&fingers ==1 && button ==0){
		EF_023_DEBUG=4;
		debug_condition_5=2;
		return 1;
	}else if(debug_condition_5==2){
	      if(fingers ==0 && button > 0){
		 EF_023_DEBUG=5;
		 debug_condition_5=3;
		}else{
		EF_023_DEBUG=5;
		debug_condition_5=5;
		}
		return 1;
	}
	else if(debug_condition_5==3 && fingers < 2 && button==0){
		debug_condition_5=4;
	}
	else if(debug_condition_5 ==4 && fingers ==1 && button==0 && w !=4){
		solution=3;
	}
	else if(debug_condition_5==5){
		EF_023_DEBUG=1;
		debug_condition_5=0;
		return 1;
	}else if (debug_condition_5 !=0 ){
		if((w ==4 && fingers==1)||(fingers > 1 && debug_condition_5 >= 3)){
			EF_023_DEBUG=1;
			debug_condition_5=0;
			return 1;
		}
		debug_condition_5=0;
	}
	else if(debug_condition_5==0 && button ==0){
		debug_condition_5=1;
	}
	
	if(fingers ==1 && button ==0){
		debug_condition_3=1;
	}
	else if(debug_condition_3==1 && fingers ==0 && button ==1){
		solution = 3;
	}else {
		debug_condition_3=0;
	}
	
	return 0;
}


/*
 * Interpret complete data packets and report absolute mode input events for
 * hardware version 1. (4 byte packets)
 */
static void elantech_report_absolute_v1(struct psmouse *psmouse)
{
	struct input_dev *dev = psmouse->dev;
	struct elantech_data *etd = psmouse->private;
	unsigned char *packet = psmouse->packet;
	int fingers;

	if (etd->fw_version_maj == 0x01) {
		/* byte 0:  D   U  p1  p2   1  p3   R   L
		   byte 1:  f   0  th  tw  x9  x8  y9  y8 */
		fingers = ((packet[1] & 0x80) >> 7) +
				((packet[1] & 0x30) >> 4);
	} else {
		/* byte 0: n1  n0  p2  p1   1  p3   R   L
		   byte 1:  0   0   0   0  x9  x8  y9  y8 */
		fingers = (packet[0] & 0xc0) >> 6;
	}

	input_report_key(dev, BTN_TOUCH, fingers != 0);

	/* byte 2: x7  x6  x5  x4  x3  x2  x1  x0
	   byte 3: y7  y6  y5  y4  y3  y2  y1  y0 */
	if (fingers) {
		input_report_abs(dev, ABS_X,
			((packet[1] & 0x0c) << 6) | packet[2]);
		input_report_abs(dev, ABS_Y, ETP_YMAX_V1 -
			(((packet[1] & 0x03) << 8) | packet[3]));
	}

	input_report_key(dev, BTN_TOOL_FINGER, fingers == 1);
	input_report_key(dev, BTN_TOOL_DOUBLETAP, fingers == 2);
	input_report_key(dev, BTN_TOOL_TRIPLETAP, fingers == 3);
	input_report_key(dev, BTN_LEFT, packet[0] & 0x01);
	input_report_key(dev, BTN_RIGHT, packet[0] & 0x02);

	
	input_sync(dev);
}

/*
 * Interpret complete data packets and report absolute mode input events for
 * hardware version 2. (6 byte packets)
 */
static void elantech_report_absolute_v2(struct psmouse *psmouse)
{
	struct input_dev *dev = psmouse->dev;
	struct elantech_data *etd = psmouse->private;
	unsigned char *packet = psmouse->packet;
	int fingers, x1, y1, x2, y2;

	/* byte 0: n1  n0   .   .   .   .   R   L */
	fingers = (packet[0] & 0xc0) >> 6;
	input_report_key(dev, BTN_TOUCH, fingers != 0);

	switch (fingers) {
	case 0:
		  input_report_abs(dev,ABS_PRESSURE,0);
		  break;
	case 1:
		/* byte 1: x15 x14 x13 x12 x11 x10 x9  x8
		   byte 2: x7  x6  x5  x4  x4  x2  x1  x0 */
		input_report_abs(dev, ABS_X, ((packet[1] & 0x0f)<< 8) | packet[2]);
		/* byte 4: y15 y14 y13 y12 y11 y10 y8  y8
		   byte 5: y7  y6  y5  y4  y3  y2  y1  y0 */
		input_report_abs(dev, ABS_Y, Public_ETP_YMAX_V2 -
			(((packet[4] & 0x0f) << 8) | packet[5]));
		if(etd->hw_version == ETF020030) {
			input_report_abs(dev,ABS_PRESSURE,60);
		}
		else {
			input_report_abs(dev,ABS_PRESSURE,
					((packet[1]&0xf0)|((packet[4]&0xf0)>>4)));
			input_report_abs(dev,ABS_TOOL_WIDTH,
					 ((((packet[0]&0x30)>>2)|((packet[3]&0x30)>>4)) & 0x0f));
		}
		
		break;

	case 2:
		/* The coordinate of each finger is reported separately with
		   a lower resolution for two finger touches */
		/* byte 0:  .   .  ay8 ax8  .   .   .   .
		   byte 1: ax7 ax6 ax5 ax4 ax3 ax2 ax1 ax0 */
		x1 = ((packet[0] & 0x10) << 4) | packet[1];
		/* byte 2: ay7 ay6 ay5 ay4 ay3 ay2 ay1 ay0 */
		y1 = Public_ETP_2FT_YMAX - (((packet[0] & 0x20) << 3) | packet[2]);
		/* byte 3:  .   .  by8 bx8  .   .   .   .
		   byte 4: bx7 bx6 bx5 bx4 bx3 bx2 bx1 bx0 */
		x2 = ((packet[3] & 0x10) << 4) | packet[4];
		/* byte 5: by7 by8 by5 by4 by3 by2 by1 by0 */
		y2 = Public_ETP_2FT_YMAX - (((packet[3] & 0x20) << 3) | packet[5]);
		
		/* For compatibility with the proprietary X Elantech driver
		   report both coordinates as hat coordinates */
		input_report_abs(dev, ABS_HAT0X, x1);
		input_report_abs(dev, ABS_HAT0Y, y1);
		input_report_abs(dev, ABS_HAT1X, x2);
		input_report_abs(dev, ABS_HAT1Y, y2);
		input_report_abs(dev,ABS_PRESSURE,60);
		break;
	case 3:
		/* byte 1: x15 x14 x13 x12 x11 x10 x9  x8
		   byte 2: x7  x6  x5  x4  x4  x2  x1  x0 */
		input_report_abs(dev, ABS_HAT2X, ((packet[1] & 0x0f)<< 8) | packet[2]);
		/* byte 4: y15 y14 y13 y12 y11 y10 y8  y8
		   byte 5: y7  y6  y5  y4  y3  y2  y1  y0 */
		input_report_abs(dev, ABS_HAT2Y, Public_ETP_YMAX_V2 -
			(((packet[4] & 0x0f) << 8) | packet[5]));
			
		if(etd->hw_version == ETF020030) {
			input_report_abs(dev,ABS_PRESSURE,60);
		}
		else {
			input_report_abs(dev,ABS_PRESSURE,
					((packet[1]&0xf0)|((packet[4]&0xf0)>>4)));
		}
		
		break;
	}
	
	input_report_key(dev, BTN_TOOL_FINGER, fingers == 1);
	input_report_key(dev, BTN_TOOL_DOUBLETAP, fingers == 2);
	input_report_key(dev, BTN_TOOL_TRIPLETAP, fingers == 3);
	input_report_key(dev, BTN_LEFT, ((packet[0] & 0x03) == 1));
	input_report_key(dev, BTN_RIGHT,((packet[0] & 0x03) == 2));
	input_report_key(dev, BTN_MIDDLE,((packet[0] & 0x03) == 3));
	
	switch (etd->hw_version) {
		case ETF0208:
			    if(etd->fw_version_min_sp != 0x00)
				break;
		case ETF0403:		
			    input_report_key(dev, BTN_1, packet[3] & 0x80); /*MKY*/
			    input_report_key(dev, BTN_2, packet[3] & 0x40); /*VF*/
		break;
	}

	input_sync(dev);
}

static psmouse_ret_t elantech_process_byte(struct psmouse *psmouse)
{
	struct elantech_data *etd = psmouse->private;
	unsigned char *packet = psmouse->packet;
	unsigned char bit7,bit6,bit5,bit4,bit3,bit2,bit1,bit0;
	unsigned char SA,B,C,D;
	static char SA_O=0,B_O=0,C_O=0,D_O=0;	
	unsigned char SA1,A1,B1,SB1,C1,D1;
	static char SA1_O=0,A1_O=0,B1_O=0,SB1_O=0,C1_O=0,D1_O=0;
	unsigned int fingers,button,w1;
	int C_B,C_C,C_D,Debug;

	if (psmouse->pktcnt < psmouse->pktsize)
		return PSMOUSE_GOOD_DATA;
	
	switch (etd->hw_version) {
	case ETF020022:
	case ETF020600:
		
			SA = packet[0];
			B  = packet[1];
			C  = packet[2];
			D  = packet[3];
			if(SA == SA_O && B == B_O && C == C_O && D == D_O)
				goto the_same_data;
		      
			if(!(packet[0] & 0x08)) //chech SA byte 
				goto shift_4byte_data;

			bit7 = (packet[1] & 0x80) >> 7;
			bit6 = (packet[1] & 0x40) >> 6;
			bit5 = (packet[1] & 0x20) >> 5;
			bit4 = (packet[1] & 0x10) >> 4;
			bit3 = (packet[1] & 0x08) >> 3;
			bit2 = (packet[1] & 0x04) >> 2;
			bit1 = (packet[1] & 0x02) >> 1;
			bit0 = (packet[1] & 0x01) ;
			C_B = (bit7 + bit6 + bit5 + bit4 + bit3 + bit2 + bit1 + bit0)%2;

			if((packet[0]&0x10) && (C_B == 1))	//chech B byte
				goto shift_4byte_data;
			else if(!(packet[0]&0x10) && (C_B == 0))
				goto shift_4byte_data;

			bit7 = (packet[2] & 0x80) >> 7;
			bit6 = (packet[2] & 0x40) >> 6;
			bit5 = (packet[2] & 0x20) >> 5;
			bit4 = (packet[2] & 0x10) >> 4;
			bit3 = (packet[2] & 0x08) >> 3;
			bit2 = (packet[2] & 0x04) >> 2;
			bit1 = (packet[2] & 0x02) >> 1;
			bit0 = (packet[2] & 0x01) ;
			C_C = (bit7 + bit6 + bit5 + bit4 + bit3 + bit2 + bit1 + bit0)%2;

			if((packet[0]&0x20) && (C_C == 1))	//chech C byte
				goto shift_4byte_data;
			else if(!(packet[0]&0x20) && (C_C == 0))
				goto shift_4byte_data;

		
			bit7 = (packet[3] & 0x80) >> 7;
			bit6 = (packet[3] & 0x40) >> 6;
			bit5 = (packet[3] & 0x20) >> 5;
			bit4 = (packet[3] & 0x10) >> 4;
			bit3 = (packet[3] & 0x08) >> 3;
			bit2 = (packet[3] & 0x04) >> 2;
			bit1 = (packet[3] & 0x02) >> 1;
			bit0 = (packet[3] & 0x01) ;
			C_D = (bit7 + bit6 + bit5 + bit4 + bit3 + bit2 + bit1 + bit0)%2;

			if((packet[0]&0x04) && (C_D == 1))	//chech D byte
				goto shift_4byte_data;
			else if(!(packet[0]&0x04) && (C_D == 0))
				goto shift_4byte_data;
			elantech_report_absolute_v1(psmouse);
			SA_O = SA;
			B_O = B;
			C_O = C;
			D_O = D;
			break;

	case ETF020030:
			SA1= packet[0];
			A1 = packet[1];
			B1 = packet[2];
			SB1= packet[3];
			C1 = packet[4];
			D1 = packet[5];
			
			if(SA1 == SA1_O && A1 == A1_O && B1 == B1_O && 
			   SB1 == SB1_O && C1 == C1_O && D1 == D1_O) {
				goto the_same_data;
			}
				
			if( (((SA1 & 0x3C) != 0x3C) && ((SA1 & 0xC0) != 0x80))  || // check Byte 1
			    (((SA1 & 0x0C) != 0x0C) && ((SA1 & 0xC0) == 0x80))  || // check Byte 1
			    (((SA1 & 0xC0) != 0x80) && (( A1 & 0xF0) != 0x00))  || // check Byte 2
			    (((SB1 & 0x3E) != 0x38) && ((SA1 & 0xC0) != 0x80))  || // check Byte 4
			    (((SB1 & 0x0E) != 0x08) && ((SA1 & 0xC0) == 0x80))  || // check Byte 4
			    (((SA1 & 0xC0) != 0x80) && (( C1 & 0xF0) != 0x00))) {// check Byte5
		 		goto shift_6byte_data;
			}
		
			elantech_report_absolute_v2(psmouse);
			SA_O = SA1;
			A1_O = A1;
			B1_O = B1;
			SB1_O = SB1;
			C1_O = C1;
			D1_O = D1;
			break;
	case ETF0208:
			if(etd->fw_version_min_sp == 0x00 ) {
				SA1= packet[0];
				A1 = packet[1];
				B1 = packet[2];
				SB1= packet[3];
				C1 = packet[4];
				D1 = packet[5];
			
				if( ((SA1 & 0x0C) != 0x04) || // check Byte 1
				    ((SB1 & 0x0f) != 0x02) ) { // check Byte  4
					goto shift_6byte_data;
				}
			
				fingers = ((packet[0] & 0xC0) >> 6);
				button = ((packet[3]&0x01 << 2) | (packet[0] & 0x03));
				w1 = (((packet[0]&0x30)>>2)|((packet[3]&0x30)>>4)) & 0x0f;
				Debug = ETF0208_DEBUG_SOLUTION(fingers,button,w1);
				if(Debug){
					psmouse->resetafter = psmouse->out_of_sync+1;
					return PSMOUSE_BAD_DATA;
				}
			
				if(SA1 == SA1_O && A1 == A1_O && B1 == B1_O && 
				  SB1 == SB1_O && C1 == C1_O && D1 == D1_O) {
					goto the_same_data;
				}
				elantech_report_absolute_v2(psmouse);
				SA_O = SA1;
				A1_O = A1;
				B1_O = B1;
				SB1_O = SB1;
				C1_O = C1;
				D1_O = D1;
				break;
			}
	case ETF020B00:
	case ETF0402:
	case ETF0401:
	case ETF0403:
			SA1= packet[0];
			A1 = packet[1];
			B1 = packet[2];
			SB1= packet[3];
			C1 = packet[4];
			D1 = packet[5];
			
			if(SA1 == SA1_O && A1 == A1_O && B1 == B1_O && 
			   SB1 == SB1_O && C1 == C1_O && D1 == D1_O) {
				goto the_same_data;
			}
			
			if( ((SA1 & 0x0C) != 0x04) || // check Byte 1
			    ((SB1 & 0x0f) != 0x02) ) { // check Byte  4
				goto shift_6byte_data;
			}
		
			elantech_report_absolute_v2(psmouse);
			SA_O = SA1;
			A1_O = A1;
			B1_O = B1;
			SB1_O = SB1;
			C1_O = C1;
			D1_O = D1;
			break;
	}

 the_same_data:
	return PSMOUSE_FULL_PACKET;
	
 shift_4byte_data:
	packet[0] = B;
	packet[1] = C;
	packet[2] = D;
	psmouse->pktcnt = psmouse->pktcnt-1;
	return PSMOUSE_GOOD_DATA;
	
 shift_6byte_data:	
	packet[0] = A1;
	packet[1] = B1;
	packet[2] = SB1;
	packet[3] = C1;
	packet[4] = D1;
	psmouse->pktcnt = psmouse->pktcnt - 1;
	return PSMOUSE_GOOD_DATA;

}



/*
 * Put the touchpad into absolute mode
 */
 
static int elantech_set_absolute_mode(struct psmouse *psmouse)
{
	struct elantech_data *etd = psmouse->private;
	unsigned char val;
	int tries = ETP_READ_BACK_TRIES;
	int rc = 0;

	switch (etd->hw_version) {
	case ETF020022:
	case ETF020600:
			etd->reg_10 = 0x14;
			etd->reg_11 = 0x8b;
			if (elantech_write_reg(psmouse, 0x10, etd->reg_10) ||
				elantech_write_reg(psmouse, 0x11, etd->reg_11)) {
				rc = -1;
			}
			break;
	case ETF020030:
	case ETF0208:
	case ETF020B00:
	case ETF0402:
	case ETF0401:
	case ETF0403:
		/* Windows driver values */
			if(etd->hw_version == ETF020030){
				etd->reg_10 = 0x54;
				etd->reg_11 = 0x8a;
			}
			else {
				etd->reg_10 = 0xc4;
				etd->reg_11 = 0x8a;
			}
		
		 
			if (elantech_write_reg(psmouse, 0x10, etd->reg_10) ||
			    elantech_write_reg(psmouse, 0x11, etd->reg_11) ) {
			      rc = -1;
			      break;
			}
			/*
			 * Read back reg 0x10. The touchpad is probably initalising
			 * and not ready until we read back the value we just wrote.
			 */
			do {
				rc = elantech_read_reg(psmouse, 0x10, &val);
				if (rc == 0)
					break;
				tries--;
				elantech_debug("elantech.c: retrying read (%d).\n",
					tries);
				msleep(ETP_READ_BACK_DELAY);
			} while (tries > 0);
			
			if (rc)
				pr_err("elantech.c: failed to read back register 0x10.\n");
			break;
	}

	if (rc)
		pr_err("elantech.c: failed to initialise registers.\n");

	return rc;
}

/*
 * Set the appropriate event bits for the input subsystem
 */
static void elantech_set_input_params(struct psmouse *psmouse)
{
	struct input_dev *dev = psmouse->dev;
	struct elantech_data *etd = psmouse->private;
	int i;
	unsigned int x_max,y_max;
	unsigned int twoft_xmax,twoft_ymax;
	unsigned char param[3];
	

	__set_bit(EV_KEY, dev->evbit);
	__set_bit(EV_ABS, dev->evbit);

	__set_bit(BTN_LEFT, dev->keybit);
	__set_bit(BTN_RIGHT, dev->keybit);
	__set_bit(BTN_MIDDLE, dev->keybit);

	__set_bit(BTN_TOUCH, dev->keybit);
	__set_bit(BTN_TOOL_FINGER, dev->keybit);
	__set_bit(BTN_TOOL_DOUBLETAP, dev->keybit);
	__set_bit(BTN_TOOL_TRIPLETAP, dev->keybit);

	switch (etd->hw_version) {
		case ETF020022:
		case ETF020600:
			input_set_abs_params(dev, ABS_X, ETP_XMIN_V1, ETP_XMAX_V1, 0, 0);
			input_set_abs_params(dev, ABS_Y, ETP_YMIN_V1, ETP_YMAX_V1, 0, 0);
			break;
		case ETF0208:
				if(etd->fw_version_min_sp != 0x00){
				   i=1;
				   goto count_edge;
				}
		case ETF020B00:			
		case ETF020030:
	fix_edge:	
			Public_ETP_YMAX_V2 = ETP_YMAX_V2;
			Public_ETP_2FT_YMAX = ETP_2FT_YMAX;
			input_set_abs_params(dev, ABS_X, ETP_XMIN_V2, ETP_XMAX_V2, 0, 0);
			input_set_abs_params(dev, ABS_Y, ETP_YMIN_V2, ETP_YMAX_V2, 0, 0);
			input_set_abs_params(dev, ABS_HAT0X, ETP_2FT_XMIN, ETP_2FT_XMAX, 0, 0);
			input_set_abs_params(dev, ABS_HAT0Y, ETP_2FT_YMIN, ETP_2FT_YMAX, 0, 0);
			input_set_abs_params(dev, ABS_HAT1X, ETP_2FT_XMIN, ETP_2FT_XMAX, 0, 0);
			input_set_abs_params(dev, ABS_HAT1Y, ETP_2FT_YMIN, ETP_2FT_YMAX, 0, 0);
			input_set_abs_params(dev, ABS_HAT2X, ETP_XMIN_V2, ETP_XMAX_V2, 0, 0);
			input_set_abs_params(dev, ABS_HAT2Y, ETP_YMIN_V2, ETP_YMAX_V2, 0, 0);
			input_set_abs_params(dev, ABS_PRESSURE, 0, 255, 0, 0);
			input_set_abs_params(dev, ABS_TOOL_WIDTH,0,16,0,0);
			__set_bit(BTN_0, dev->keybit);
			__set_bit(BTN_1, dev->keybit);
			break;
    		case ETF0402:
     		case ETF0401:
		case ETF0403:
			i=2;
	count_edge:
			if (synaptics_send_cmd(psmouse, ETP_CAPABILITIES_QUERY, param)) {
				pr_err("elantech.c: failed to query capabiliities.\n");
				goto fix_edge;
			}
			
			x_max=param[1];
			x_max=(x_max-i)*64;
			y_max=param[2];	
			y_max=(y_max-i)*64;
			
			
			twoft_xmax=(x_max-i)*64/4;			
			twoft_ymax=(y_max-i)*64/4;
			Public_ETP_YMAX_V2 = y_max;
			Public_ETP_2FT_YMAX = twoft_ymax;

			input_set_abs_params(dev, ABS_X, ETP_XMIN_V2, x_max, 0, 0);
			input_set_abs_params(dev, ABS_Y, ETP_YMIN_V2, y_max, 0, 0);
			input_set_abs_params(dev, ABS_HAT0X, ETP_2FT_XMIN, twoft_xmax, 0, 0);
			input_set_abs_params(dev, ABS_HAT0Y, ETP_2FT_YMIN, twoft_ymax, 0, 0);
			input_set_abs_params(dev, ABS_HAT1X, ETP_2FT_XMIN, twoft_xmax, 0, 0);
			input_set_abs_params(dev, ABS_HAT1Y, ETP_2FT_YMIN, twoft_ymax, 0, 0);
			input_set_abs_params(dev, ABS_HAT2X, ETP_XMIN_V2, x_max, 0, 0);
			input_set_abs_params(dev, ABS_HAT2Y, ETP_YMIN_V2, y_max, 0, 0);
			input_set_abs_params(dev, ABS_PRESSURE, 0, 255, 0, 0);
			input_set_abs_params(dev, ABS_TOOL_WIDTH,0,16,0,0);
			__set_bit(BTN_0, dev->keybit);
			__set_bit(BTN_1, dev->keybit);
				
				
      }
}


struct elantech_attr_data {
	size_t		field_offset;
	unsigned char	reg;
};

/*
 * Display a register value by reading a sysfs entry
 */
static ssize_t elantech_show_int_attr(struct psmouse *psmouse, void *data,
					char *buf)
{
	struct elantech_data *etd = psmouse->private;
	struct elantech_attr_data *attr = data;
	unsigned char *reg = (unsigned char *) etd + attr->field_offset;
	int rc = 0;

	if (attr->reg)
		rc = elantech_read_reg(psmouse, attr->reg, reg);

	return sprintf(buf, "0x%02x\n", (attr->reg && rc) ? -1 : *reg);
}

/*
 * Write a register value by writing a sysfs entry
 */
static ssize_t elantech_set_int_attr(struct psmouse *psmouse,
				     void *data, const char *buf, size_t count)
{
	struct elantech_data *etd = psmouse->private;
	struct elantech_attr_data *attr = data;
	unsigned char *reg = (unsigned char *) etd + attr->field_offset;
	unsigned long value;
	int err;

	err = strict_strtoul(buf, 16, &value);
	if (err)
		return err;

	if (value > 0xff)
		return -EINVAL;

	/* Do we need to preserve some bits for version 2 hardware too? */
	if (etd->hw_version < ETF020030) {
		if (attr->reg == 0x10)
			/* Force absolute mode always on */
			value |= ETP_R10_ABSOLUTE_MODE;
		else if (attr->reg == 0x11)
			/* Force 4 byte mode always on */
			value |= ETP_R11_4_BYTE_MODE;
	}

	if (!attr->reg || elantech_write_reg(psmouse, attr->reg, value) == 0)
		*reg = value;

	return count;
}

#define ELANTECH_INT_ATTR(_name, _register)				\
	static struct elantech_attr_data elantech_attr_##_name = {	\
		.field_offset = offsetof(struct elantech_data, _name),	\
		.reg = _register,					\
	};								\
	PSMOUSE_DEFINE_ATTR(_name, S_IWUSR | S_IRUGO,			\
			    &elantech_attr_##_name,			\
			    elantech_show_int_attr,			\
			    elantech_set_int_attr)

ELANTECH_INT_ATTR(reg_10, 0x10);
ELANTECH_INT_ATTR(reg_11, 0x11);
ELANTECH_INT_ATTR(reg_20, 0x20);
ELANTECH_INT_ATTR(reg_21, 0x21);
ELANTECH_INT_ATTR(reg_22, 0x22);
ELANTECH_INT_ATTR(reg_23, 0x23);
ELANTECH_INT_ATTR(reg_24, 0x24);
ELANTECH_INT_ATTR(reg_25, 0x25);
ELANTECH_INT_ATTR(reg_26, 0x26);
ELANTECH_INT_ATTR(debug, 0);
ELANTECH_INT_ATTR(paritycheck, 0);

static struct attribute *elantech_attrs[] = {
	&psmouse_attr_reg_10.dattr.attr,
	&psmouse_attr_reg_11.dattr.attr,
	&psmouse_attr_reg_20.dattr.attr,
	&psmouse_attr_reg_21.dattr.attr,
	&psmouse_attr_reg_22.dattr.attr,
	&psmouse_attr_reg_23.dattr.attr,
	&psmouse_attr_reg_24.dattr.attr,
	&psmouse_attr_reg_25.dattr.attr,
	&psmouse_attr_reg_26.dattr.attr,
	&psmouse_attr_debug.dattr.attr,
	&psmouse_attr_paritycheck.dattr.attr,
	NULL
};

static struct attribute_group elantech_attr_group = {
	.attrs = elantech_attrs,
};


/*
 * Use magic knock to detect Elantech touchpad
 */

int elantech_detect(struct psmouse *psmouse, int set_properties)
{
	struct ps2dev *ps2dev = &psmouse->ps2dev;
	unsigned char param[3];
	printk(KERN_INFO "2.6.2X-Elan-touchpad-2009-05-21\n");
	ps2_command(&psmouse->ps2dev, NULL, PSMOUSE_CMD_DISABLE);
	
	param[0] = param[1] = param[2] = 0;
	
	if(!ps2_command(&psmouse->ps2dev, param, PSMOUSE_CMD_RESET_BAT)) {
		printk(KERN_INFO "elantech.c: PSMOUSE_CMD_RESET_BAT  param[0]=%x param[1]=%x param[2]=%x\n"
				  ,param[0],param[1],param[2]);
	}
	
	    
	if (ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE) ||
	    ps2_command(ps2dev,  NULL, PSMOUSE_CMD_SETSCALE11) ||
	    ps2_command(ps2dev,  NULL, PSMOUSE_CMD_SETSCALE11) ||
	    ps2_command(ps2dev,  NULL, PSMOUSE_CMD_SETSCALE11) ||
	    ps2_command(ps2dev, param, PSMOUSE_CMD_GETINFO)) {
		pr_err("elantech.c: sending Elantech magic knock failed.\n");
		return -1;
	}

	/*
	 * Report this in case there are Elantech models that use a different
	 * set of magic numbers
	 */
	if (param[0] != 0x3c || param[1] != 0x03 || param[2] != 0xc8) {
		pr_info("elantech.c: unexpected magic knock result 0x%02x, 0x%02x, 0x%02x.\n",
			param[0], param[1], param[2]);
		return -1;
	}

	if (set_properties) {
		psmouse->vendor = "Elantech";
		
	}

	return 0;
}

/*
 * Clean up sysfs entries when disconnecting
 */
static void elantech_disconnect(struct psmouse *psmouse)
{
	sysfs_remove_group(&psmouse->ps2dev.serio->dev.kobj,
			   &elantech_attr_group);
	kfree(psmouse->private);
	psmouse->private = NULL;
}

/*
 * Put the touchpad back into absolute mode when reconnecting
 */
static int elantech_reconnect(struct psmouse *psmouse)
{
	if (elantech_detect(psmouse, 0))
		return -1;

	if (elantech_set_absolute_mode(psmouse)) {
		pr_err("elantech.c: failed to put touchpad back into absolute mode.\n");
		return -1;
	}
	return 0;
}

static int elantech_reconnect_ETF0208(struct psmouse *psmouse)
{
	struct ps2dev *ps2dev = &psmouse->ps2dev;
	int i;
	int err,debug_err;
	err=i8042_command(NULL,I8042_CMD_KBD_DISABLE);// tom +

	if(!err)
		printk(KERN_INFO"i8042_command I8042_CMD_KBD_DISABLE OK\n");
	
	switch(EF_023_DEBUG){
	
		case 1:
				ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x04);
				debug_err=elantech_write_reg(psmouse, 0x60, 0x00);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x00);
				EF_023_DEBUG=0;
				
				break;

		case 2:
				ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x04);	
				debug_err=elantech_write_reg(psmouse, 0x1F, 0x00);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x00);
				EF_023_DEBUG=0;
				break;
		case 3:				
				ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x04);	
				debug_err=elantech_write_reg(psmouse, 0x13, 0x08);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x00);
				EF_023_DEBUG=0;
				break;	
		case 4:				
				ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x04);	
				debug_err=elantech_write_reg(psmouse, 0x06, 0x07);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x00);
				EF_023_DEBUG=0;
				break;	
		case 5:				
				ps2_command(ps2dev,  NULL, PSMOUSE_CMD_DISABLE);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x04);	
				debug_err=elantech_write_reg(psmouse, 0x06, 0x17);
				debug_err=elantech_write_reg(psmouse, 0x12, 0x00);
				EF_023_DEBUG=0;	
				break;	


		default:

			for(i=0;i<3;i++){
				printk(KERN_INFO"elantech_reconnect\n");
				if (!elantech_detect(psmouse, 0)){		
					if (!elantech_init(psmouse))
			 	 	break;
				}
				else
					continue;
			}
	
			if(i>=3){
				printk(KERN_INFO"elantech_reconnect fail\n");
				if(!err)
					err=i8042_command(NULL,I8042_CMD_KBD_ENABLE);// tom +	
					if(!err)
					printk(KERN_INFO"i8042_command I8042_CMD_KBD_ENABLE OK\n");
					return -1;
			}
			break;

	}
	if(!err)
		err=i8042_command(NULL,I8042_CMD_KBD_ENABLE);// tom +	
	if(!err)
		printk(KERN_INFO"i8042_command I8042_CMD_KBD_ENABLE OK\n");
	return 0;
}
/*
 * Initialize the touchpad and create sysfs entries
 */
int elantech_init(struct psmouse *psmouse)
{
	struct elantech_data *etd;
	int i, error;
	unsigned char param[3];
	

	etd = kzalloc(sizeof(struct elantech_data), GFP_KERNEL);
	psmouse->private = etd;
	if (!etd)
		return -1;

	etd->parity[0] = 1;
	for (i = 1; i < 256; i++)
		etd->parity[i] = etd->parity[i & (i - 1)] ^ 1;

	/*
	 * Find out what version hardware this is
	 */
	if (synaptics_send_cmd(psmouse, ETP_FW_VERSION_QUERY, param)) {
		pr_err("elantech.c: failed to query firmware version.\n");
		goto init_fail;
	}
	pr_info("elantech.c: Elantech version query result 0x%02x, 0x%02x, 0x%02x.\n",
		param[0], param[1], param[2]);
	etd->fw_version_maj = param[0];
	etd->fw_version_min = param[1];
	etd->fw_version_min_sp = param [2];

	
	/*
	 * Assume every version greater than this is Elan Touch Pad style
	 * hardware with 6 byte packets
	 */
	 
	if (etd->fw_version_maj == 0x02 && etd->fw_version_min == 0x00 && 
	    etd->fw_version_min_sp == 0x22) {
		etd->hw_version = ETF020022;
		psmouse->name = "ETF020022";
	}
	else if (etd->fw_version_maj == 0x02 && etd->fw_version_min == 0x06 && 
		etd->fw_version_min_sp == 0x00) {
	        etd->hw_version = ETF020600;
		psmouse->name = "ETF020600";
	}
	else if (etd->fw_version_maj == 0x02 && etd->fw_version_min == 0x00 && 
		etd->fw_version_min_sp == 0x30) {
	        etd->hw_version = ETF020030;
		psmouse->name = "ETF020030";
	}
	else if (etd->fw_version_maj == 0x02 && etd->fw_version_min == 0x08 ) {
		 etd->hw_version = ETF0208; 
		 psmouse->name = "ETF0208";
	}
	else if (etd->fw_version_maj == 0x02 && etd->fw_version_min == 0x0B &&
		etd->fw_version_min_sp == 0x00) {
		etd->hw_version = ETF020B00;
		psmouse->name = "ETF020B00";
	}
	else if (etd->fw_version_maj == 0x04 && etd->fw_version_min == 0x02) {
	        etd->hw_version = ETF0402;
		psmouse->name = "ETF0402";
	}
	else if (etd->fw_version_maj == 0x04 && etd->fw_version_min == 0x01) {
	        etd->hw_version = ETF0401;
		psmouse->name = "ETF0401";
	}
	else if (etd->fw_version_maj == 0x04 && etd->fw_version_min == 0x03) {
	        etd->hw_version = ETF0403;
		psmouse->name = "ETF0403";
	}
	else {
              printk(KERN_DEBUG "elantech -- don't find firmware version\n");
	      goto init_fail;
	}
	pr_info("elantech.c: assuming hardware version %d, firmware version 0x%02x.0x%02x.0x%02x\n",
	etd->hw_version, etd->fw_version_maj, etd->fw_version_min,etd->fw_version_min_sp);
	
	if (elantech_set_absolute_mode(psmouse)) {
		pr_err("elantech.c: failed to put touchpad into absolute mode.\n");
		goto init_fail;
	}

	elantech_set_input_params(psmouse);

	error = sysfs_create_group(&psmouse->ps2dev.serio->dev.kobj,
				   &elantech_attr_group);
	if (error) {
		pr_err("elantech.c: failed to create sysfs attributes, error: %d.\n",
			error);
		goto init_fail;
	}

	psmouse->protocol_handler = elantech_process_byte;
	psmouse->disconnect = elantech_disconnect;
	if( etd->hw_version == ETF0208 && 
	    etd->fw_version_min_sp == 0x00){
		EF_023_DEBUG = 0;
		psmouse->reconnect =elantech_reconnect_ETF0208;
	}else {
		psmouse->reconnect = elantech_reconnect;
	}
	psmouse->pktsize = etd->hw_version > ETF020600 ? 6 : 4;

	return 0;

 init_fail:
	kfree(etd);
	return -1;
}

